import aiorun
import logging
from telethon import TelegramClient, errors

from tgpy import app, Config
from tgpy.console import console
from tgpy.handlers import add_handlers
from tgpy.hooks import HookType, Hook
from tgpy.utils import SESSION_FILENAME, create_config_dirs

log = logging.getLogger(__name__)


def create_client():
    client = TelegramClient(str(SESSION_FILENAME), app.config.api_id, app.config.api_hash)
    client.parse_mode = 'html'
    return client


async def start_client():
    await app.client.start(
        phone=lambda: console.input('| Please enter your phone number: '),
        code_callback=lambda: console.input('| Please enter the code you received: '),
        password=lambda: console.input('| Please enter your 2FA password: ', password=True),
    )


async def initial_setup():
    console.print('[bold on bright_cyan] Welcome to TGPy ')
    console.print('Starting setup...')
    console.print()
    console.print('[bold black on white] Step 1 of 2 ')
    console.print(
        '│ TGPy uses Telegram API, so you\'ll need to register your Telegram app.\n'
        '│  [cyan]1.[/] Go to https://my.telegram.org\n'
        '│  [cyan]2.[/] Login with your Telegram account\n'
        '│  [cyan]3.[/] Go to "API development tools"\n'
        '│  [cyan]4.[/] Create your app. Choose any app title and short_title. You can leave other fields empty.\n'
        '│ You will get api_id and api_hash.'
    )
    success = False
    while not success:
        app.config.api_id = console.input('│ Please enter api_id: ')
        app.config.api_hash = console.input('│ ...and api_hash: ')
        try:
            app.client = create_client()
            console.print()
            console.print('[bold black on white] Step 2 of 2 ')
            console.print('│ Now login to Telegram.')
            await app.client.connect()
            await start_client()
            success = True
        except (errors.ApiIdInvalidError, errors.ApiIdPublishedFloodError, ValueError):
            console.print('│ [bold on red]Incorrect api_id/api_hash, try again')
        finally:
            if app.client:
                await app.client.disconnect()
    console.print('│ Login successful!')
    app.config.save()


async def run_client():
    log.info('Starting TGPy...')
    await start_client()
    log.info('[bold]TGPy is running!', extra={'markup': True})
    await Hook.run_hooks(HookType.onstart)
    await app.client.run_until_disconnected()


async def _main():
    create_config_dirs()

    app.config = Config.load()
    if not (app.config.api_id and app.config.api_hash):
        await initial_setup()
    else:
        app.client = create_client()

    add_handlers()
    await run_client()


def main():
    aiorun.run(_main(), stop_on_unhandled_errors=True)



import importlib.metadata
import shlex
from enum import Enum
from pathlib import Path
from subprocess import Popen, PIPE

import appdirs
import yaml
from pydantic import ValidationError
from yaml.representer import SafeRepresenter

DATA_DIR = Path(appdirs.user_config_dir('tgpy'))
HOOKS_DIR = DATA_DIR / 'hooks'
WORKDIR = DATA_DIR / 'workdir'
CONFIG_FILENAME = DATA_DIR / 'config.yml'
SESSION_FILENAME = DATA_DIR / 'TGPy.session'


def create_config_dirs():
    DATA_DIR.mkdir(exist_ok=True)
    HOOKS_DIR.mkdir(exist_ok=True)
    WORKDIR.mkdir(exist_ok=True)


def _multiline_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')


def _enum_presenter(dumper, data):
    value = data.value
    if isinstance(value, str):
        return dumper.represent_str(value)
    if isinstance(value, int):
        return dumper.represent_int(value)
    if isinstance(value, bool):
        return dumper.represent_bool(value)
    raise ValueError


class yaml_multiline_str(str):
    pass


SafeRepresenter.add_representer(yaml_multiline_str, _multiline_presenter)
SafeRepresenter.add_multi_representer(Enum, _enum_presenter)


class RunCmdException(Exception):
    def __init__(self, process: Popen):
        self.process = process

    def __str__(self):
        return f'Command {shlex.join(self.process.args)} exited with code {self.process.returncode}'


def run_cmd(args: list[str]):
    proc = Popen(args, stdout=PIPE)
    output, _ = proc.communicate()
    if proc.returncode:
        raise RunCmdException(proc)
    return output.decode('utf-8').strip()


def installed_as_package():
    try:
        importlib.metadata.version('tgpy')
        return True
    except importlib.metadata.PackageNotFoundError:
        return False


def get_version():
    if installed_as_package():
        return importlib.metadata.version('tgpy')

    try:
        return 'git@' + run_cmd(['git', 'rev-parse', '--short', 'HEAD'])
    except RunCmdException:
        pass

    return 'unknown'


__all__ = ['DATA_DIR', 'HOOKS_DIR', 'WORKDIR', 'CONFIG_FILENAME', 'SESSION_FILENAME',
           'yaml_multiline_str', 'run_cmd', 'get_version', 'create_config_dirs',
           'installed_as_package']



from typing import Callable

from telethon.tl.custom import Message

from tgpy import message_design


def _handle_errors(func: Callable):
    async def result(message: Message):
        # noinspection PyBroadException
        try:
            await func(message)
        except Exception:
            await message_design.send_error(message.chat_id)

    return result


def outgoing_messages_filter(m: Message):
    return m.out and not m.forward and not m.via_bot



from telethon.errors import MessageIdInvalidError
from telethon.tl.custom import Message

from tgpy import app
from tgpy import message_design
from tgpy.run_code.meval import meval
from tgpy.run_code.utils import Output, convert_result, filename_prefix, format_traceback
from tgpy.run_code.variables import variables


def get_kwargs(include_orig=True):
    return list(variables.keys()) + ['ctx', 'msg', 'print', 'client'] + ['orig'] if include_orig else []


async def eval_message(code: str, message: Message, uses_orig=False) -> None:
    await message_design.edit_message(message, code, 'Running...')

    output = Output()

    variables['ctx'].msg = message

    kwargs = {}
    if uses_orig:
        orig = await message.get_reply_message()
        kwargs['orig'] = orig

    # noinspection PyBroadException
    try:
        new_variables, result = await meval(
            code,
            f'{filename_prefix}message/{message.chat_id}/{message.id}',
            globals(),
            variables,
            client=app.client,
            msg=message,
            ctx=variables['ctx'],
            print=output.print,
            **kwargs,
        )
    except Exception:
        result = 'Error occurred'
        exc = ''.join(format_traceback())
    else:
        variables.update(new_variables)
        result = convert_result(result)
        exc = ''

    try:
        await message_design.edit_message(message, code, result, traceback=exc, output=output.text)
    except MessageIdInvalidError:
        pass


from tgpy.run_code import builtin_functions



import logging

from rich.logging import RichHandler
from telethon import TelegramClient

from tgpy.app_config import Config
from tgpy.console import console


__version__ = "0.2.3"

logging.basicConfig(level=logging.INFO, format='%(message)s', datefmt="[%X]", handlers=[RichHandler()])


class App:
    config: Config = None
    client: TelegramClient = None


app = App()



from typing import Optional

import yaml
from pydantic import BaseModel, ValidationError

from tgpy.utils import CONFIG_FILENAME


class Config(BaseModel):
    api_id: Optional[int]
    api_hash: Optional[str]

    @classmethod
    def load(cls):
        try:
            with open(CONFIG_FILENAME) as file:
                config_dict = yaml.safe_load(file)
        except FileNotFoundError:
            config_dict = {}
        try:
            config = cls(**config_dict)
        except ValidationError:
            config = cls()
        return config

    def save(self):
        with open(CONFIG_FILENAME, 'w') as file:
            yaml.safe_dump(self.dict(), file)



import getpass
import os
import socket
import sys
from textwrap import dedent
from typing import Optional

import datetime as dt

from tgpy.hooks import Hook, HookType, delete_hook_file, get_sorted_hooks
from tgpy.message_design import get_code
from tgpy.run_code import variables
from tgpy.run_code.utils import Context, filename_prefix, save_function_to_variables
from tgpy.utils import run_cmd, get_version, installed_as_package, RunCmdException

variables['ctx'] = ctx = Context()


@save_function_to_variables
def ping():
    return f'Pong!\n' \
           f'Running on {getpass.getuser()}@{socket.gethostname()}\n' \
           f'Version: {get_version()}'


@save_function_to_variables
def restart(msg: Optional[str] = 'Restarted successfully'):
    hook_code = dedent(f'''
        from tgpy.message_design import edit_message, get_code
        msg = await client.get_messages({ctx.msg.chat_id}, ids={ctx.msg.id})
        await edit_message(msg, get_code(msg), '{msg}')
    ''')
    hook = Hook(
        name='__restart_message',
        type=HookType.onstart,
        once=True,
        save_locals=False,
        code=hook_code,
        origin=f'{filename_prefix}restart_message',
        datetime=dt.datetime.fromtimestamp(0),
    )
    hook.save()
    os.execl(sys.executable, sys.executable, '-m', 'tgpy', *sys.argv[1:])


@save_function_to_variables
def update():
    if installed_as_package():
        update_args = [sys.executable, '-m', 'pip', 'install', '-U', 'tgpy']
        try:
            run_cmd(update_args)
        except RunCmdException:
            run_cmd(update_args + ['--user'])
    else:
        run_cmd(['git', 'pull'])
    restart(f'Updated successfully! Current version: {get_version()}')


class HooksObject:
    async def add(self, name: str, code: Optional[str] = None) -> str:
        if code is None:
            original = await ctx.msg.get_reply_message()
            if original is None:
                return 'Use this function in reply to a message'
            code = get_code(original)
            origin = f'{filename_prefix}message/{original.chat_id}/{original.id}'
        else:
            origin = f'{filename_prefix}message/{ctx.msg.chat_id}/{ctx.msg.id}'

        hook = Hook(
            name=name,
            type=HookType.onstart,
            once=False,
            save_locals=True,
            code=code,
            origin=origin,
            datetime=dt.datetime.now(),
        )
        hook.save()
        return dedent(f'''
            Added hook {name!r}.
            The hook will be executed every time TGPy starts.
        ''')

    def remove(self, name) -> str:
        try:
            delete_hook_file(name)
        except FileNotFoundError:
            return f'No hook named {name!r}.'
        return f'Removed hook {name!r}.'

    def __str__(self):
        lst = '\n'.join(f'{idx + 1}. {hook.name}' for idx, hook in enumerate(get_sorted_hooks()))
        if not lst:
            return dedent('''
                You have no hooks.
                Learn about hooks at https://tgpy.tmat.me/hooks.
            ''')
        return dedent('''
            Your hooks:
            {}
            
            Change hooks with `hooks.add(name)` and `hooks.remove(name)`.
            Learn more at https://tgpy.tmat.me/hooks.
        ''').format(lst)

    def __iter__(self):
        return (hook.name for hook in get_sorted_hooks())


variables['hooks'] = hooks = HooksObject()



from telethon import events
from telethon.tl.custom import Message
from telethon.tl.types import Channel

from tgpy import app, message_design
from tgpy.handlers.utils import _handle_errors, outgoing_messages_filter
from tgpy.run_code import eval_message, get_kwargs
from tgpy.run_code.parse_code import parse_code


async def handle_message(message: Message) -> None:
    raw_text = message.raw_text

    if not raw_text:
        return

    if message.text.startswith('//') and message.text[2:].strip():
        await message.edit(message.text[2:])
        return

    locals_ = get_kwargs()

    res = parse_code(raw_text, locals_)
    if not res.is_code:
        return

    await eval_message(raw_text, message, uses_orig=res.uses_orig)


@events.register(events.NewMessage(func=outgoing_messages_filter))
@_handle_errors
async def on_new_message(event: events.NewMessage.Event) -> None:
    await handle_message(event.message)


@events.register(events.MessageEdited(func=outgoing_messages_filter))
@_handle_errors
async def on_message_edited(event: events.NewMessage.Event) -> None:
    if isinstance(event.message.chat, Channel) and event.message.chat.broadcast:
        return
    code = message_design.get_code(event.message)
    if not code:
        await handle_message(event.message)
        return
    await eval_message(code, event.message, uses_orig=parse_code(code, get_kwargs()).uses_orig)


@events.register(events.NewMessage(pattern='(?i)^(cancel|сфтсуд)$', func=outgoing_messages_filter))
async def cancel(message: Message):
    prev = await message.get_reply_message()
    if not prev:
        async for msg in app.client.iter_messages(message.chat_id, max_id=message.id, limit=10):
            if msg.out and message_design.get_code(msg):
                prev = msg
                break
        else:
            return
    # noinspection PyBroadException
    try:
        await prev.edit(message_design.get_code(prev))
    except Exception:
        pass
    else:
        await message.delete()


def add_handlers():
    app.client.add_event_handler(on_new_message)
    app.client.add_event_handler(on_message_edited)
    app.client.add_event_handler(cancel)



import sys
import traceback
from typing import Optional

from telethon.tl import TLObject
from telethon.tl.custom import Message

from tgpy.run_code.variables import variables


class Context:
    msg: Optional[Message] = None

    def __str__(self):
        return f'<Context(msg)>'


class Output:
    text = ''

    def print(self, *values, sep=' ', end='\n', file=None, flush=True):
        if file:
            print(*values, sep=sep, end=end, file=file, flush=flush)
        else:
            self.text += sep.join(str(val) for val in values) + end


def convert_result(result):
    if isinstance(result, TLObject):
        result = result.stringify()

    return result


filename_prefix = 'tgpy://'


def format_traceback():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    exc_traceback = exc_traceback.tb_next.tb_next
    return traceback.format_exception(exc_type, exc_value, exc_traceback)


def save_function_to_variables(func):
    variables[func.__name__] = func
    return func




# forked from https://pypi.org/project/meval/

import ast
import inspect
from collections import deque
from importlib.abc import SourceLoader
from importlib.util import spec_from_loader, module_from_spec
from types import CodeType
from typing import Any, Iterator


def shallow_walk(node) -> Iterator:
    # Like ast.walk, but ignoring function definitions:
    # Recursively yield all descendant nodes except for function definitions in the tree starting at node
    queue = deque([node])
    while queue:
        node = queue.popleft()
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            continue
        queue.extend(ast.iter_child_nodes(node))
        yield node


class MevalLoader(SourceLoader):
    source: bytes
    code: CodeType
    filename: str

    def __init__(self, source: str, code: CodeType, filename: str):
        self.source = source.encode('utf-8')
        self.code = code
        self.filename = filename

    def is_package(self, _):
        return False

    def get_filename(self, _):
        return self.filename

    def get_data(self, _):
        return self.source

    def get_code(self, _):
        return self.code


async def meval(str_code: str, filename: str, globs: dict, saved_variables: dict, **kwargs) -> (dict, Any):
    kwargs.update(saved_variables)

    # Restore globals later
    globs = globs.copy()

    # This code saves __name__ and __package into a kwarg passed to the function.
    # It is set before the users code runs to make sure relative imports work
    global_args = '_globs'
    while global_args in globs.keys():
        # Make sure there's no name collision, just keep prepending _s
        global_args = '_' + global_args
    kwargs[global_args] = {}
    for glob in ['__name__', '__package__']:
        # Copy data to args we are sending
        kwargs[global_args][glob] = globs[glob]

    root = ast.parse(str_code, 'exec')

    ret_name = '_ret'
    ok = False
    while True:
        if ret_name in globs.keys():
            ret_name = '_' + ret_name
            continue
        for node in ast.walk(root):
            if isinstance(node, ast.Name) and node.id == ret_name:
                ret_name = '_' + ret_name
                break
            ok = True
        if ok:
            break

    code = root.body
    if not code:
        return {}, None

    # globals().update(**<global_args>)
    glob_copy = ast.Expr(
        ast.Call(
            func=ast.Attribute(
                value=ast.Call(
                    func=ast.Name(id='globals', ctx=ast.Load()),
                    args=[],
                    keywords=[]
                ),
                attr='update',
                ctx=ast.Load()
            ),
            args=[],
            keywords=[
                ast.keyword(arg=None, value=ast.Name(id=global_args, ctx=ast.Load()))
            ]
        )
    )
    ast.fix_missing_locations(glob_copy)
    code.insert(0, glob_copy)

    # _ret = []
    ret_decl = ast.Assign(
        targets=[ast.Name(id=ret_name, ctx=ast.Store())],
        value=ast.List(elts=[], ctx=ast.Load())
    )
    ast.fix_missing_locations(ret_decl)
    code.insert(1, ret_decl)

    # __import__('builtins').locals()
    get_locals = ast.Call(
        func=ast.Attribute(
            value=ast.Call(
                func=ast.Name(id='__import__', ctx=ast.Load()),
                args=[ast.Constant(value='builtins')],
                keywords=[]
            ),
            attr='locals',
            ctx=ast.Load()
        ),
        args=[],
        keywords=[]
    )

    if not any(isinstance(node, ast.Return) for node in shallow_walk(root)):
        for i in range(len(code)):
            if not isinstance(code[i], ast.Expr) \
                    or i != len(code) - 1 and isinstance(code[i].value, ast.Call):
                continue

            # replace ... with _ret.append(...)
            code[i] = ast.copy_location(
                ast.Expr(
                    ast.Call(
                        func=ast.Attribute(
                            value=ast.Name(id=ret_name, ctx=ast.Load()),
                            attr='append',
                            ctx=ast.Load()
                        ),
                        args=[code[i].value],
                        keywords=[]
                    )
                ),
                code[-1]
            )
    else:
        for node in shallow_walk(root):
            if not isinstance(node, ast.Return):
                continue

            # replace return ... with return (__import__('builtins').locals(), [...])
            node.value = ast.Tuple(
                elts=[
                    get_locals,
                    ast.List(elts=[node.value or ast.Constant(value='None')], ctx=ast.Load())
                ],
                ctx=ast.Load()
            )

    # return (__import__('builtins').locals(), _ret)
    code.append(
        ast.copy_location(
            ast.Return(
                value=ast.Tuple(
                    elts=[
                        get_locals,
                        ast.Name(id=ret_name, ctx=ast.Load())
                    ],
                    ctx=ast.Load()
                )
            ),
            code[-1]
        )
    )

    args = []
    for a in list(map(lambda x: ast.arg(x, None), kwargs.keys())):
        ast.fix_missing_locations(a)
        args += [a]
    args = ast.arguments(args=[], vararg=None, kwonlyargs=args, kwarg=None, defaults=[],
                         kw_defaults=[None for i in range(len(args))])
    args.posonlyargs = []
    fun = ast.AsyncFunctionDef(name='tmp', args=args, body=code, decorator_list=[], returns=None)
    ast.fix_missing_locations(fun)
    mod = ast.Module(body=[fun], type_ignores=[])

    # print(ast.unparse(mod))
    comp = compile(mod, filename, 'exec')
    loader = MevalLoader(str_code, comp, filename)
    py_module = module_from_spec(spec_from_loader(filename, loader, origin=filename))
    loader.exec_module(py_module)

    new_locs, ret = await getattr(py_module, 'tmp')(**kwargs)
    for loc in list(new_locs):
        if (loc in globs or loc in kwargs or loc == ret_name) and loc not in saved_variables:
            new_locs.pop(loc)

    ret = [await el if inspect.isawaitable(el) else el for el in ret]
    ret = [el for el in ret if el is not None]

    if len(ret) == 1:
        ret = ret[0]
    elif not ret:
        ret = None

    new_locs['_'] = ret
    return new_locs, ret



import ast


class _Result:
    is_code = False
    uses_orig = False


def _is_node_unknown_variable(node: ast.AST, locs: dict) -> bool:
    """Check if AST node is a Name or Attribute not present in locals"""
    if isinstance(node, ast.Attribute):
        return _is_node_unknown_variable(node.value, locs)
    return isinstance(node, ast.Name) and node.id not in locs


def _is_node_suspicious_binop(node: ast.AST, locs: dict) -> bool:
    """Check if AST node can be an operand of binary operation (ast.BinOp, ast.Compare, ast.BoolOp)
    with operands which do not pass _is_node_unknown_variable check, or is such operation"""
    if _is_node_unknown_variable(node, locs):
        return True
    if not isinstance(node, (ast.BoolOp, ast.BinOp, ast.Compare)):
        return False
    if isinstance(node, ast.Compare):
        return _is_node_unknown_variable(node.left, locs) and all(_is_node_unknown_variable(x, locs)
                                                                  for x in node.comparators)
    return all(_is_node_suspicious_binop(operand, locs)
               for operand in ((node.left, node.right) if isinstance(node, ast.BinOp) else node.values))


def _ignore_node(node: ast.AST, locs: dict) -> bool:
    """Check if AST node didn't seem to be meant to be code"""
    return (
        # Messages like "python", "123" or "example.com"
        isinstance(node, ast.Constant) or _is_node_unknown_variable(node, locs)
        # Messages like "-1", "+spam" and "not foo.bar"
        # `getattr(..., None) or node.value` is used here to avoid AttributeError and because in UnaryOp and Starred
        # operands are stored in different attributes ("operand" and "value" respectively)
        or isinstance(node, (ast.Starred, ast.UnaryOp)) and isinstance(getattr(node, 'operand', None) or node.value,
                                                                       (ast.Constant, ast.Name, ast.Attribute))
        # Messages like one-two, one is two, one >= two, one.b in two.c
        or _is_node_suspicious_binop(node, locs)
        # Messages like "yes, understood"
        or isinstance(node, ast.Tuple) and all(_ignore_node(elt, locs) for elt in node.elts)
    )


def parse_code(text: str, locs: dict) -> _Result:
    """Parse given text and decide should it be evaluated as Python code"""
    result = _Result()

    try:
        root = ast.parse(text, '', 'exec')
    except (SyntaxError, ValueError):
        return result

    if all(isinstance(body_item, ast.Expr) and _ignore_node(body_item.value, locs) for body_item in root.body):
        return result

    result.is_code = True

    for node in ast.walk(root):
        if isinstance(node, ast.Name) and node.id == 'orig':
            result.uses_orig = True
            return result

    return result




import sys
import traceback as tb

from telethon.tl.custom import Message
from telethon.tl.types import MessageEntityTextUrl, MessageEntityCode, MessageEntityBold

from tgpy import app

TITLE = 'TGPy>'
TITLE_URL = 'https://github.com/tm-a-t/TGPy'
ERROR_TITLE_FORMATTED = f'<b><a href="{TITLE_URL}">TGPy error&gt;</a></b>'


async def edit_message(message: Message, code: str, result, traceback: str = '', output: str = '') -> None:
    if result is None and output:
        result = output
        output = ''

    parts = [code.strip(), f'{TITLE} {str(result).strip()}']
    parts += [part for part in (output.strip(), traceback.strip()) if part]
    text = '\n\n'.join(parts)

    entities = []
    offset = 0
    for p in parts:
        entities.append(MessageEntityCode(offset, len(p)))
        offset += len(p) + 2

    entities[1].offset += len(TITLE) + 1
    entities[1].length -= len(TITLE) + 1
    entities[1:1] = [
        MessageEntityBold(len(parts[0]) + 2, len(TITLE)),
        MessageEntityTextUrl(len(parts[0]) + 2, len(TITLE), TITLE_URL),
    ]

    if len(text) > 4096:
        text = text[:4095] + '…'
    await message.edit(text, formatting_entities=entities, link_preview=False)


def get_code(message: Message) -> str:
    for e in message.entities or []:
        if isinstance(e, MessageEntityTextUrl) and e.url == TITLE_URL:
            return message.raw_text[:e.offset].strip()
    return ''


async def send_error(chat) -> None:
    exc = ''.join(tb.format_exception(*sys.exc_info()))
    if len(exc) > 4000:
        exc = exc[:4000] + '…'
    await app.client.send_message(chat, f'{ERROR_TITLE_FORMATTED}\n\n<code>{exc}</code>', link_preview=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tgpy', 'tgpy.handlers', 'tgpy.run_code']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Telethon>=1.24.0,<2.0.0',
 'aiorun>=2021.10.1,<2022.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'cryptg>=0.2.post4,<0.3',
 'pydantic>=1.8.2,<2.0.0',
 'rich>=10.16.1,<11.0.0']

entry_points = \
{'console_scripts': ['tgpy = tgpy.main:main']}

setup_kwargs = {
    'name': 'tgpy',
    'version': '0.2.3',
    'description': 'Run Python code right in your Telegram messages',
    'long_description': '![TGPy](https://raw.githubusercontent.com/tm-a-t/TGPy/master/readme_assets/TGPy.png)\n\n## Run Python code right in your Telegram messages\n\nMade with Telethon library, TGPy is a tool for evaluating expressions and Telegram API scripts.\n\n- Do Python calculations in dialogs\n- Interact with your messages and chats\n- Automate sending messages and more\n\n![Example](https://raw.githubusercontent.com/tm-a-t/TGPy/master/readme_assets/example.gif)\n\nJust send Python code to any chat, and it will be executed. Change your message to change the result.\n\n# Examples\n\nSend any of these examples to any chat to evaluate:\n\n🐍 Do Python calculations\n\n```python\nfor i in range(5):\n    print(i)\n```\n\n⏳ Delete the current message in 5 seconds\n\n```python\nimport asyncio\n\nawait asyncio.sleep(5)\nawait msg.delete()\n```\n\n↪️ Forward the message you replied to to another chat\n\n```python\norig.forward_to(\'Chat title\')\n```\n\n🖼 Send all chat profile photos to the same chat\n\n```python\nphotos = await client.get_profile_photos(msg.chat)\nmsg.reply(file=photos)\n```\n\n🔖 Define a function which forwards messages to Saved Messages with reply\n\n```python\ndef save():\n    message = ctx.msg\n    original = await message.get_reply_message()\n    await original.forward_to(\'me\')\n    return \'Saved!\'\n``` \n\n🗑 Define a function which deletes messages with reply\n\n```python\nasync def delete():\n    message = ctx.msg\n    original = await message.get_reply_message()\n    await original.delete()\n    await message.delete()\n```\n\n# Guide\n\n- [Basics](https://tgpy.tmat.me/basics/)\n- [TGPy builtins](https://tgpy.tmat.me/builtins/)\n- [Code detection](https://tgpy.tmat.me/code_detection/)\n\n# Installation\n\n1. You\'ll need Telegram API key. Register your "app" at [my.telegram.org](https://my.telegram.org) to get `api_id` and \n`api_hash`. App title and other data don\'t matter.\n\n2. Clone the repo:\n   ```shell\n   > git clone https://github.com/tm-a-t/TGPy\n   > cd TGPy\n   ```\n\n3. Create `config.py`. Enter your API data and your phone to log in:\n   ```python\n   api_id = ...\n   api_hash = ...\n   phone = ...\n   ```\n\n4. Install the requirements and run TGPy:\n   ```shell\n   > pip install -r requirements.txt\n   > python -m tgpy\n   ```\n\n5. For the first time, you\'ll need to log in with a confirmation code from Telegram.\n\nSend `ping()` to any chat to check if TGPy is running.\n\n# Credits\n\n- Thanks to [penn5](https://github.com/penn5) for [meval](https://github.com/penn5/meval)\n- Thanks to [Lonami](https://github.com/LonamiWebs) for [Telethon](https://github.com/LonamiWebs/Telethon)\n\n# License\n\nThis project is licensed under the terms of the MIT license.\n',
    'author': 'tmat',
    'author_email': 'a@tmat.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tm-a-t/TGPy/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import datetime as dt
import logging
import traceback
from enum import Enum
from pathlib import Path
from typing import Union

import yaml
from pydantic import BaseModel

from tgpy import app
from tgpy.run_code import meval
from tgpy.run_code.utils import format_traceback
from tgpy.run_code.variables import variables
from tgpy.utils import yaml_multiline_str, HOOKS_DIR

logger = logging.getLogger(__name__)


class HookType(str, Enum):
    onstart = 'onstart'


def get_hook_filename(name: Union[str, Path]) -> Path:
    name = Path(name)
    if name.suffix not in ['.yml', '.yaml']:
        name = name.with_suffix('.yml')
    return HOOKS_DIR / name


def delete_hook_file(name: Union[str, Path]):
    get_hook_filename(name).unlink()


def get_sorted_hooks(type_=None):
    hooks = []

    for file in HOOKS_DIR.iterdir():
        if file.suffix not in ['.yml', '.yaml']:
            continue
        hook_name = file.stem

        # noinspection PyBroadException
        try:
            hook = Hook.load(hook_name)
        except Exception:
            logger.error(f'Error during loading hook {hook_name!r}')
            logger.error(traceback.format_exc())
            continue
        if type_ is None or hook.type == type_:
            hooks.append(hook)

    hooks.sort(key=lambda hook: hook.datetime)
    return hooks


class Hook(BaseModel):
    name: str
    type: HookType
    once: bool
    save_locals: bool = True
    code: str
    origin: str
    datetime: dt.datetime

    @classmethod
    def load(cls, hook_name: str) -> 'Hook':
        filename = get_hook_filename(hook_name)
        with open(filename) as f:
            hook = Hook.parse_obj(yaml.safe_load(f))
        if hook.name != hook_name:
            raise ValueError(f'Invalid hook name. Expected: {hook_name!r}, found: {hook.name}')
        return hook

    def save(self):
        filename = get_hook_filename(self.name)
        data = self.dict()
        data['code'] = yaml_multiline_str(data['code'])

        with open(filename, 'w') as f:
            f.write(yaml.safe_dump(data))

    async def run(self):
        new_variables, result = await meval(
            self.code,
            self.origin,
            globals(),
            variables,
            client=app.client,
            msg=None,
            ctx=variables['ctx'],
            print=lambda *args, **kwargs: None,
        )
        if self.save_locals:
            variables.update(new_variables)
        if self.once:
            delete_hook_file(self.name)

    @classmethod
    async def run_hooks(cls, type_: HookType):
        for hook in get_sorted_hooks(type_):
            # noinspection PyBroadException
            try:
                await hook.run()
            except Exception:
                logger.error(f'Error during running hook {hook.name!r}')
                logger.error(''.join(format_traceback()))
                continue
